// *************************************************************************
//    BindTime.h
// *************************************************************************

#ifndef _BindTime_H_
#define _BindTime_H_

#include "BitVector.h"

class Cor_RuleInfo;

// *************************************************************************

enum Cor_Unify_Funct {
    COR_USE_UNIFY=0,    // Use full unify algorithm
    COR_USE_ASSIGN,     // Free Variable vs Bound, use Assignment
    COR_USE_EQ_CHK,     // Both bound, check for equality
    COR_USE_MATCH       // Use matching algorithm (one bound, other has
                        // variables within its structure) 
};

// *************************************************************************
// A class which performs a binding-time analysis of a rule. It assumes the
// head literal is using duplicate checking to generate the backtracking
// info, (which is trival otherwise).
// The analysis collects the first binding point for each variable, and the
// last literal the variable occurs in, this should be sufficient for most
// backtracking algorithms.
// *************************************************************************

   // The different types of literals
enum  Cor_BT_Literal_Kind {
    COR_BT_LIT_NORMAL=0,         // Nothing special, move along       
    COR_BT_LIT_CUT,              // This is a cut predicate
    COR_BT_LIT_SINGLE,           // This predicate is a "single-success"
    COR_BT_LIT_ARITHPRED         // A arithmetic predicate (=,<,>=, etc)
};

class Cor_BindTime {

  private:

    int               num_vars;   // Total number of unique vars in this rule
    Cor_BitVector    *head_vars;  // Variables in the head of the rule

    int               num_lits;   // Number of righthand-side literals in rule

    Cor_BitVector   **lit_vars;   // Array of variables in each literal
                                  // 0 corresponds to the left-most (first)
                                  // literal in the rule
                          // [ They are ordered according to the order
                          // given in the PredOrders array, so it
                          // is not needed after the analysis is finished. ]
   
 
    Cor_BT_Literal_Kind  *lit_kind;    // Array giving type of each literal 

      // The head literal is not considered for
      // the following (calculated) tables.

    int              *first_bind; // Array of first literal thats binds
                                  // each variable. (indexed by Var->varnum())

    int              *last_seen;  // Array of last literal thats the variable
                                  // appears in.


    void init() {
        num_vars = 0;
        num_lits = 0;
        head_vars = NULL;
        lit_vars  = NULL;
        first_bind = NULL;
        last_seen = NULL;
        lit_kind = NULL;
    }

    int CutCrossed(int bt, int src);
    int skipSingleSuccess(int bt);

  public:

    Cor_BindTime() { init(); }
   ~Cor_BindTime() { clear(); }

      // Do the actual Binding-Time-Anaylsis
    int Analyze(Cor_RuleInfo *rInfo, int *predOrders=NULL);

      // Fix order so no arithmetic predicates bind variables
    int Cor_BindTime::FixOrder(int *predOrders);

      // Set the functions to use when unifying later
      // [against a fact known to be ground (containing no variables).]
    int set_UnifyFunctions(Cor_RuleInfo *rInfo, Cor_Unify_Funct **unifiers,
            int *predOrders=NULL);

      // Get first-failure backtrack point 
      // (Used when literal fails the first time it is reached.)
    int get_firstFailure_BT_Pt(int *bt_array);

      // Get next-failure backtrack point
      // (Used when literal fails after it has succeeded a few times.)
    int get_nextFailure_BT_Pt(int *bt_array);

      // Get success backtrack point
      // (Used after entire rule has been sucessfully instantiated, to
      //  attempt to derive another fact from it.)
    int get_Success_BT_Pt();

    void clear() {

        delete head_vars;

        for (int j=0; j < num_lits; j++)
            delete lit_vars[j];
        delete [] lit_vars;

        delete [] lit_kind;
        delete [] first_bind; 
        delete [] last_seen; 

        init();
    }

      // For debugging (output goes to cout)
    void show_raw_info(int *predOrders=NULL);
};

// *************************************************************************

#endif _BindTime_H_
