
// *************************************************************************
//    SockConnectArray.C
//
//    The SockConnectArray Class - Used to implement connections
//    on TCP/IP sockets. 
//
//    A ConnectArray holds an array of Connections, of which each
//    holds an array of Users.
//     
// *************************************************************************

#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include "Identify.h"
#include "SockConnectArray.h"
#include "SockStream.h"
#include "Globals.h"
#include "ClientProtocol.h"
#include "Debug.h"

// *************************************************************************
// SockConnectArray Constructor - Just clears everything.
// *************************************************************************
Cor_SockConnectArray::Cor_SockConnectArray(unsigned short port)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    cur_conn = 0;
    num_conn = 0;
    nxt_conn = 0;
    max_conn = MAX_CONNECTIONS;

    for (int i = 0; i < MAX_CONNECTIONS; i++)
        connArray[i] = (Cor_Connection *) NULL;

    memset(&select_msk, '\0', sizeof(select_msk));

#ifdef    DO_EXPLAIN
    expl_serv_strm = NULL;
#endif

    serv_buf = new Cor_sockbuf(port);
    if (!serv_buf || ! serv_buf->valid()) {
        cerr << "Error: Could not start Coral server.\n";
        exit(0);
    }

    FD_SET(serv_buf->socketfd(), &select_msk);

#ifdef    DO_EXPLAIN

      // Explain server socket
    unsigned short eport = EXPLAIN_PORT;
    Cor_sockbuf *tmp_buf = new Cor_sockbuf(eport);
    if (!tmp_buf || ! tmp_buf->valid())
        return;

    expl_serv_strm = new CoralStream(tmp_buf);
    if (!expl_serv_strm)
        return;

    FD_SET(tmp_buf->socketfd(), &select_msk);

#endif

    // This next message is helpful for the test suite programs.
    cout << "[" << getpid() << "]: Coral Server started on port "
         <<  port << ".\n";
    cout << flush;
}

// *************************************************************************
// SockConnectArray Destructor - Closes the server streams, then closes
//    each connection in the array.
// *************************************************************************
Cor_SockConnectArray::~Cor_SockConnectArray()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    delete serv_buf;

#ifdef    DO_EXPLAIN
      // Close explain server
    delete expl_serv_strm;
#endif

      // Close all connections
    for (int i = 0; i < MAX_CONNECTIONS; i++)
        if (connArray[i])
            delete connArray[i];
}

// *************************************************************************
// NextConn() -
//    This function finds the next available connection with data pending
//    and switches to it.  This function also checks for connection
//    requests on the server socket and accepts them. If more than one
//    connection has data waiting to be read, repeatedly calling this
//    function will switch between them in a round-robin fashion.
//
//    Return: The return value is the number of connections with data,
//        or < 0 if an error occurs.
//
// *************************************************************************
int Cor_SockConnectArray::NextConn()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int    num = 0;
    fd_set tmp_msk;

    do {
          // Duplicate the select mask.
        memcpy(&tmp_msk, &select_msk, sizeof(fd_set));


        // Find next connection (round-robin) with data pending
        // Block until we recieve a request to connect or data
        // If a connection request is recieved, connect it and
        // continue waiting for data on any port.

#ifdef HPUX
        if ((num = select(FD_SETSIZE,(int *) &tmp_msk, 0, 0,
                            (struct timeval *) NULL)) <= 0) {
#else
        if ((num = select(FD_SETSIZE,&tmp_msk, 0, 0,
                            (struct timeval *) NULL)) <= 0) {
#endif
            perror("Coral: Select Error ");
            num = 0;
            continue;
            //exit(0);
        }
   
        if (FD_ISSET(serv_buf->socketfd(), &tmp_msk)){
  
              // Connection request 
            NewConnection();
            num--;
        }

#ifdef    DO_EXPLAIN

        Cor_sockbuf *tmp_buf = (Cor_sockbuf *) expl_serv_strm->rdbuf();
        if (FD_ISSET(tmp_buf->socketfd(), &tmp_msk)){

              // Connection request (for Explain)
            ExplainConnection();
            num--;
        }
#endif

    } while(!num);

      // If only 1 port has data, start at the one you're currently at.
      // Otherwise, start at the next (round-robin)
    if (num != 1)
        cur_conn = (cur_conn + 1) % MAX_CONNECTIONS;

    int prv_conn = cur_conn;

    do {

        Cor_Connection *conn;
        if ((conn = connArray[cur_conn])) {

            Cor_sockbuf *tmp_buf = (Cor_sockbuf *) conn->conn_strm->rdbuf();

            if (FD_ISSET(tmp_buf->socketfd(), &tmp_msk)){

                  // Make this connection the current
                conn->useConnection();

                  // Return the number of connections with data (why not?)
                return(num);
            }
        }

        cur_conn = (cur_conn + 1) % MAX_CONNECTIONS;

    } while(cur_conn != prv_conn);

    cerr << "Could not locate a socket with data?!\n";

    return(-1);
}

// *************************************************************************
// NewConnection() -
//    This function is callled when a connection request is recieved on the
//    server port.  This function decides if the new connection should be
//    allowed and accepts or refuses it. 
//
//    Return: The return value is the connection id of the new connection,
//        or < 0 if an error occurs or the connection is refused.
//
// *************************************************************************
int Cor_SockConnectArray::NewConnection()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int this_conn;

    if (!serv_buf) 
        return(-1);

      // Check for at max connections
    if (num_conn >= max_conn){
        serv_buf->RefuseConnection();
        return(-1);
    }

      // The max_conn value can be changed dynamically
    if (nxt_conn >= max_conn)
        nxt_conn = 0;

    this_conn = nxt_conn;
    while (connArray[this_conn] && (this_conn < max_conn))
        this_conn++;

      // Place to start looking for next connection
    nxt_conn = this_conn + 1; 

    Cor_sockbuf *client_buf = serv_buf->AcceptConnection();
    CoralStream *client_strm = new CoralStream(client_buf);
    connArray[this_conn] = new Cor_Connection(client_strm);
    connArray[this_conn]->conn_id = this_conn;
    connArray[this_conn]->sock_fd = client_buf->socketfd();

    FD_SET(client_buf->socketfd(), &select_msk);
    num_conn++;

    return(this_conn);
}

// *************************************************************************
// DeleteConnection() -
//    This function deletes a connection from the SockConnectArray object.
//
// *************************************************************************
void Cor_SockConnectArray::DeleteConnection(int conn_indx)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (! connArray[conn_indx])
        return;

    Cor_sockbuf *tmp_buf = (Cor_sockbuf *)
            (connArray[conn_indx])->conn_strm->rdbuf();

    if (tmp_buf->socketfd() >= 0)
        FD_CLR(connArray[conn_indx]->sock_fd, &select_msk);

    delete connArray[conn_indx];
    connArray[conn_indx] = (Cor_Connection *) NULL;

    num_conn--;

      // Keep track of lowest available connection
    nxt_conn = (conn_indx < nxt_conn) ? conn_indx : nxt_conn;
}

// *************************************************************************
// ExplainConnection() -
//    This function is callled when a connection request is recieved on the
//    explain port.  This function decides if the connection attempt is
//    valid and accepts or refuses it.
//
//    Return: The return value is zero if everything is okay
//        or < 0 if an error occurs or the connection is refused.
//
// *************************************************************************
int Cor_SockConnectArray::ExplainConnection()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
#ifdef    DO_EXPLAIN

    char buff[500];
    int c_id=0, u_id=0, pnt=0;

    if (!expl_serv_strm) 
        return(-1);

    Cor_sockbuf *tmp_buf = (Cor_sockbuf*) expl_serv_strm->rdbuf();
    Cor_sockbuf *tmp_sock = tmp_buf->AcceptConnection();
    CoralStream *in = new CoralStream(tmp_sock);

    if (! in->gets(buff, 500)) {
        return(-1);
    }

    sscanf(buff, "%d %d %n", &c_id, &u_id, &pnt); 
   
    if (u_id){ 
        for (int i=pnt; buff[i]; i++)
            if ((buff[i] == '\r') || (buff[i] == '\n'))
                buff[i] = '\0';
    } else
        buff[pnt] = '\0';

    Cor_Connection *conn = connArray[c_id];
    Cor_User *user = NULL;

    if ((! conn) || (! (user = conn->userArray[u_id])) ||
            strcmp(user->password,&buff[pnt]) || 
            (user->Cor_exEnv.explain_strm != NULL) || 
            ! (user->Cor_exEnv.GlobalRelationOptions & REL_EXPLAIN)){

        int sd = tmp_sock->socketfd();
        char *mess = "Connection refused by Coral Server.\n";
        write(sd,mess,strlen(mess));

          // Disconnect
        shutdown(sd,2);
        close(sd);
        return(0);
    }

      // They check out, set them up to recieve explain info
    if (user) {
        user->explain_strm = in;
        user->Cor_exEnv.explain_strm = in;
        // And inform client (client blocks waiting for a message)
        *(user->Cor_exEnv.out_strm) << endl
                                    << CP_MESG << "Explain Connected.\n";
        *(user->Cor_exEnv.out_strm) << flush;
        *(user->Cor_exEnv.explain_strm) << "coral.\n" << flush;
    }

#endif    // DO_EXPLAIN

    return(0);
}

// *************************************************************************
